{-# LANGUAGE FlexibleInstances, DeriveDataTypeable #-}

module Main where

import Data.Generics.Zipper
import Data.Maybe
import Data.Data
import Data.List
import System.IO	

{----------- A few notes ------------
comp1 >- cert1: where cert is a parallel computations means that the output of comp1
		 will feed all the computations of cert1, therefore the output type of comp1
		 MUST be the same of ALL the input types of the subprocesses of cert1

ParallelNode ProcessingTree (ProcessingList): In this node of the ProcessingTree, 
		 the type of the first ProcessingTree must ALWAYS be ProcessComp i.e., a 
		 parallel computation must ALWAYS have as root a component capable of aggregating
		 all the results of the subprocesses, which MUST be the same. The >| and >|>

When the combinator +> is used to create a certification, the sub tree is validated,
		therefore the attr 'typeCheck' for ProcessCert is always True
-------------------------------------}

{----------- AG Supporting functions ------------}
-- Gives the n'th child
(.$) :: Zipper a -> Int -> Zipper a
z .$ 1 = fromJust (down' z)
z .$ n = fromJust (right ( z.$(n-1) ))

-- Tests if z is the n'th sibling
(.|) :: Zipper a -> Int -> Bool
z .| 1 = case (left z) of
			Nothing -> True
			_ -> False
z .| n = case (left z) of
			Nothing -> False
			Just x ->  z .| (n-1)
		
parent = fromJust.up

toAG t = toZipper $ RootTree t

constructor :: Typeable a => Zipper a -> String
constructor a = case ( getHole a :: Maybe ProcessingTree ) of
				   Just (RootTree     _    ) -> "RootTree"
				   Just (ProcessCert  _    ) -> "ProcessCert"
				   Just (ProcessComp _ _ _ ) -> "ProcessComp"
				   Just (SequenceNode _ _  ) -> "SequenceNode"
				   Just (ParallelNode _ _  ) -> "ParallelNode"
				   Just (Input )             -> error "**** Constructor 'Input' is NOT a Tree node! ****"
				   Nothing -> case ( getHole a :: Maybe ProcessingList ) of
				   				Just (ProcessingList _ _   ) -> "ProcessingList"
								Just (ProcessingListNode _ ) -> "ProcessingListNode"
				   				Nothing -> error "**** Bad input on 'construtor' ****"

lexemmeArg :: Typeable a => Zipper a -> Arg
lexemmeArg z = case ( getHole z :: Maybe Arg) of
					Just arg -> arg
					Nothing  -> error "**** Bad input on 'lexemmeArg'! ****"

lexemmeInputList :: Typeable a => Zipper a -> InputList
lexemmeInputList z = case ( getHole z :: Maybe Component) of
						Just (Component _ inp _ _) -> inp
						Nothing                    -> error "**** Bad input on 'lexemmeInput'! ****"

lexemmeOutputList :: Typeable a => Zipper a -> OutputList
lexemmeOutputList z = case ( getHole z :: Maybe Component) of
						Just (Component _ _ out _) -> out
						Nothing                    -> error "**** Bad input on 'lexemmeInput'! ****"

lexemmeBashCall :: Typeable a => Zipper a -> BashCall
lexemmeBashCall z = case ( getHole z :: Maybe Component) of
						Just (Component _ _ _ call) -> call
						Nothing                     -> error "**** Bad input on 'lexemmeBashCall'! ****"

{----------- AG attributes ------------}
run :: Typeable a => Zipper a -> String
run z = case constructor z of
				"RootTree" -> if (typeCheck z) then (toScript z 1)
				              else error "**** The types on the type do not match! ****"
				otherwise  -> error "**** Bad input on Attr 'run'! ****"

toScript :: Typeable a => Zipper a -> Int -> String
toScript z i = case constructor z of
				"RootTree"           -> let h1 =  "#!/usr/bin/perl"                       
				                            h2 = "use strict;"                            
				                            h3 = "use warnings;"                          
				                            h4 = "use IO::CaptureOutput qw/capture_exec/;"
				                            h5 = "use IO::File;"                          
				                            h6 = "$| = 1;"                                
				                            h7 = "my $stdout0 = $ARGV[0]; # This is actually stdin" ++ "\n"
				                            go = toScript (z.$1) i
				                        in  unlines [h1,h2,h3,h4,h5,h6,h7,go]
				"ProcessComp"        -> let bash_call = (lexemmeBashCall $ z.$1)++" "++(lexemmeArg $ z.$2)++" "++(lexemmeArg $ z.$3)
				                            n         = show i												  
				                            line1     = "my $cmd"++n++" = \""++bash_call++"\";"
				                            line2_1   = "my ($stdout"++n++", $stderr"++n++", $success"++n++", $exit_code"++n++") = " 
				                            line2_2   = "capture_exec( $cmd"++n++" . \"<<END\\n\" . $stdout"++show(i-1)++" . \"\\nEND\" );"
				                            line3     = "if ($?) { die \"**** The process $cmd"++n++" does not exist! ****\"; }"
				                            line4     = "if (not $success"++n++") { die \"**** The process $cmd"++n++" failed with msg: $stderr"++n++" ! ****\"; }"
				                            all_lines = unlines [line1,line2_1++line2_2,line3,line4]
				                            last_line = "\n"++"print $stdout"++n++";" 
				                        in  if   (isRightMost z) -- It's the last one (or the only one)
										    then all_lines++last_line
											else all_lines
				"ProcessCert"        -> toScript (z.$1) i
				"SequenceNode"       -> let tree1 = toScript (z.$1) i
				                            nodes_right = (numberNodes $ z.$1) + i 
				                            tree2 = toScript (z.$2) nodes_right
				                        in  concat [tree1,"\n",tree2]
				"ParallelNode"       -> let component      = parallelNodeHelper (z.$2) i (numberNodes $ z.$1)
				                            pList          = toScript (z.$1) i
				                            init_help      = "#Before starting a parallel computationg we default the stdout to 0"
				                            initialization = "$stdout0 = $stdout"++show (i-1)++";\n"
				                            -- The parallel computation starts the script if it is leftmost
											-- otherwise it needs to be initialized (re-channel the stdouts)
									    in  if (isLeftMost z) then pList++component
									                          else unlines [init_help,initialization,pList,component]
				"ProcessingList"     -> let n      = show i
				                            line1  = "my $handle"++n++" = new IO::File();"            
				                            line2  = "my $pid"++n++" = open($handle"++n++", \"-|\");" 
				                            line3  = "if ($pid"++n++" == 0) {"                        
				                            middle = toScript (z.$1) 1
				                            out    = "print $stdout"++(show $ numberNodes $ z.$1)++";"
				                            line4  = "exit;\n}\n"                                     
				                            all_lines = unlines [line1,line2,line3,middle++out,line4]
				                        in  all_lines ++ (toScript (z.$2) (i+1))
				"ProcessingListNode" -> let n      = show i
				                            line1  = "my $handle"++n++" = new IO::File();"            
				                            line2  = "my $pid"++n++" = open($handle"++n++", \"-|\");" 
				                            line3  = "if ($pid"++n++" == 0) {"                        
				                            middle = toScript (z.$1) 1
				                            out    = "print $stdout"++(show $ numberNodes $ z.$1)++";"											
				                            line4  = "exit;\n}\n"                                       
				                            all_lines = unlines [line1,line2,line3,middle++out,line4]
				                        in  all_lines
				otherwise  -> error "**** Bad input on Attr 'toScript'! ****"

parallelNodeHelper z i n = let grab_help    = "# Lets grab all the results of the parallel processes"
                               grab_outputs = "my $stdout"++show (i+n-1)++" = "++draw_handlers n (i+n-1)++";"
                               component    = toScript z (i+n)
		                   in  unlines [grab_help,grab_outputs,component]
                           where draw_handlers 1 n = "<$handle"++show n++">"
                                 draw_handlers i n = (draw_handlers (i-1) (n-1))++". \" \" . <$handle"++show n++">"
							
numberNodes :: Typeable a => Zipper a -> Int
numberNodes z = case constructor z of
				"ProcessComp"        -> 1
				"ProcessCert"        -> (numberNodes $ z.$1)
				"SequenceNode"       -> (numberNodes $ z.$1) + (numberNodes $ z.$2)
				"ParallelNode"       -> (numberNodes $ z.$1) + (numberNodes $ z.$2)
				"ProcessingList"     -> 1 + (numberNodes $ z.$2)
				"ProcessingListNode" -> 1				
				otherwise            -> error "**** Bad input on Attr 'numberNodes'! ****"

isLeftMost :: Typeable a => Zipper a -> Bool
isLeftMost z = case constructor z of
				"RootTree"           -> True
				_                    -> case (left z) of
				                         Nothing -> isLeftMost $ parent z
				                         Just _  -> False

isRightMost :: Typeable a => Zipper a -> Bool
isRightMost z = case constructor z of
				"RootTree"           -> True
				_                    -> case (right z) of
				                         Nothing -> isRightMost $ parent z
				                         Just _  -> False

typeCheck :: Typeable a => Zipper a -> Bool
typeCheck z = case constructor z of
				"RootTree"           -> typeCheck $ z.$1
				"ProcessComp"        -> True
				"ProcessCert"        -> True
				"SequenceNode"       -> ((output $ z.$1) == (input $ z.$2)) && (typeCheck $ z.$1) && (typeCheck $ z.$2)
				"ParallelNode"       -> (typeCheck $ z.$2) && ((input $ z.$2) == (output $ z.$1))
				"ProcessingList"     -> (input $ z.$1) == (input $ z.$2) && (output $ z.$1) == (output $ z.$2)
				"ProcessingListNode" -> True
				otherwise            -> error "**** Bad input on Attr 'typeCheck'! ****"

input :: Typeable a => Zipper a -> Language
input z = case constructor z of
			"RootTree"           -> input $ z.$1
			"ProcessCert"        -> input $ z.$1
			"ProcessComp"        -> getLanguageByArg (lexemmeArg $ z.$2) (lexemmeInputList $ z.$1)
			"SequenceNode"       -> input $ z.$1
			"ParallelNode"       -> input $ z.$1
			-- Important note: parallel processing requires that all the the processes have the
			-- same input. Therefore, the input of parallel processing is the input of one of
			-- the processes (they MUST all be the same). Despite, we have chosen a safe approach
			-- here: we only give the input AFTER we validate that inputs of the processes
			-- are ALL the same, otherwise we generate an error.
			"ProcessingList"     -> if (typeCheck z) then input $ z.$1
		                                             else error "**** The types of a parallel computation are NOT uniform! ****"
			"ProcessingListNode" -> input $ z.$1
			otherwise            -> error "**** Bad input on Attr 'input'! ****"

output :: Typeable a => Zipper a -> Language
output z = case constructor z of
			"RootTree"     -> output $ z.$1
			"ProcessCert"  -> output $ z.$1
			"ProcessComp"  -> getLanguageByArg (lexemmeArg $ z.$3) (lexemmeOutputList $ z.$1)
			"SequenceNode" -> output $ z.$2
			"ParallelNode" -> output $ z.$2
			-- Check out the note on the 'input' attr for ProcessingList. The same applies here.
			"ProcessingList"      -> if (typeCheck z) then output $ z.$1
			                                     	  else error "**** The types of a parallel computation are NOT uniform! ****"
			"ProcessingListNode"  -> output $ z.$1 
			otherwise             -> error "**** Bad input on Attr 'output'! ****"

{----------- Combinators and Tree DataTypes ------------}
data Language = Java     -- .java
			  | C_Source -- .c
			  | C_Header -- .h
		      | Cpp      -- .cpp
		      | Haskell  -- .cpp
		      | XML      -- .xml
		      | Report   -- Report XML
			  deriving (Show, Data, Typeable, Eq, Ord)

type Arg      = String
type Name     = String
type BashCall = String
type InputList  = [(Arg, Language)]
type OutputList = [(Arg, Language)]

data Component = Component Name InputList OutputList BashCall
				   deriving (Show, Data, Typeable)

-- Certification output must always be Report
-- and the input must always be ony ONE language
data Certification = Certification Name ProcessingTree
				   deriving (Show, Data, Typeable)

data ProcessingList = ProcessingList ProcessingTree ProcessingList
					| ProcessingListNode ProcessingTree
					deriving (Show, Data, Typeable)

(|:|) :: ProcessingTree -> ProcessingList -> ProcessingList
tree |:| list = ProcessingList tree list

getLanguageByArg :: Arg -> InputList -> Language
getLanguageByArg a l = let aux = (filter ((==a) . fst) l)
		               in  if length aux < 1 then error "**** Bad arguments on 'getLanguageByArg'! ****"
					                          else snd $ head aux

data ProcessingTree = RootTree ProcessingTree
					| SequenceNode ProcessingTree ProcessingTree -- SequenceNode: first > second
					| ParallelNode ProcessingList ProcessingTree -- The last argument must ALWAYS be a (ProcessComp Component)
					| ProcessCert  ProcessingTree
					| ProcessComp  Component Arg Arg		
					| Input
					deriving (Show, Typeable, Data)

{----------- Sequence Combinators ------------}
infixl 6 >-
class Sequencing node where
	(>-) :: ProcessingTree -> node -> ProcessingTree

instance Sequencing Certification where
	Input >- (Certification _ t) = ProcessCert t
	tree  >- (Certification _ t) = SequenceNode tree (ProcessCert t)

instance Sequencing (Component,Arg,Arg) where
	Input >- (comp,f1,f2) = ProcessComp comp f1 f2
	tree  >- (comp,f1,f2) = SequenceNode tree (ProcessComp comp f1 f2)

instance Sequencing (ProcessingTree) where
	tree1 >- tree2 = SequenceNode tree1 tree2

{----------- Parallel Combinators ------------}
infixr 4 >|
(>|) :: ProcessingTree -> ProcessingTree -> ProcessingTree
tree1 >| (ParallelNode listTree comp) = ParallelNode (tree1 |:| listTree) comp

infixr 5 >|>
(>|>) :: ProcessingTree -> (Component,Arg,Arg) -> ProcessingTree
tree >|> (comp,f1,f2) = ParallelNode (ProcessingListNode tree) (ProcessComp comp f1 f2)

{----------- Create Certification Combinator ------------}
infixr 3 +>
(+>) :: ProcessingTree -> String -> Certification
tree +> name = let ag = toZipper $ RootTree tree
               -- In this setting AGs (and their attributes)
			   -- are first order, so let's use them!!
               in  if (output ag == Report && (typeCheck $ ag.$1))
                   then Certification name tree
				   else error $ "**** The types int the processing tree '"++name++"' do not match! ****"

{----------- Process Everything ------------}
infixr 2 #>>
class Processing input where
	(#>>) :: input -> Char -> String

instance Processing ProcessingTree where
	t #>> 't' = show $ typeCheck $ toZipper $ RootTree t
	t #>> 's' = run $ toAG t
	_ #>> a   = "**** Invalid option for '#>>'. Try 's' or 't'! ****"

instance Processing Certification where
	(Certification _ t) #>> 't' = show $ typeCheck $ toAG t
	(Certification _ t) #>> 's' = show $ run $ toAG t	
	_ #>> a   = "**** Invalid option for '#>>'. Try 's' or 't'! ****"

{----------- Tests ------------}
-- We have created a set of Components and Certifications to test our setting
-- To make it easy to alter the parameters, we have created components and can
-- take and output virtually all types, to make it easier to play with
-- the type checking system

-- Also, all components are related to the same process, called 'script+1.pl'
-- This script behaves as expected, i.e., it takes information from STDIN and 
-- outputs from STDOUT, and the arguments are negleted. This script takes
-- a set of numbers thourgh STDIN and increments their sum by 1.
-- This way, is easy to see if the output of the script generated is correct,
-- being it the resuts of how many sequences of parallelizations the user desires.

-- Please remember that, although all the components read and write though STDIN and
-- STDOUT, the main script, always saved in 'script.pl', takes input as arguments

-- As an example, currently the script generated is the result of 'cert2' that feeds
-- a parallel computation. In you call this script as in './script.pl 1', 'cert2' will
-- increment it to 2, and 2 will feed all the parallel computations, so each one will 
-- return 3. 'comp1', the aggregator of the parallel computation will take all the
-- outputs through STDIN (3 3 3) and will increment their sum by 1, returning 10.

allPossible = [("-j",Java),("-c",C_Source),("-h",C_Header),("-cpp",Cpp),("-h",Haskell),("-x",XML),("-r",Report)]

comp1 = Component "Comp1" allPossible allPossible "./script+1.pl"
comp2 = Component "Comp2" allPossible allPossible "./script+1.pl"
comp3 = Component "Comp3" allPossible allPossible "./script+1.pl"

cert1 = Certification "Cert1" (ProcessComp comp1 "-x" "-r")
cert2 = Certification "Cert2" (ProcessComp comp2 "-c" "-r")
cert3 = Certification "Cert3" (Input >- cert2 >- (comp1,"-r","-j") >- (comp1,"-j","-h") >- (comp2,"-h","-r") >- cert1)

sequence1 = Input >- (comp2, "-h", "-x") >- (comp1,"-x","-r")

sequence2= Input >- cert1 >- parallel >- (comp1,"-x","-c") >- cert2 >- (comp2,"-r","-c")

sequence3 = Input >- cert2 >- (comp1,"-r","-j") >- (comp1,"-j","-c") >- cert3 >- (comp2,"-r","-x") >- cert1

parallel  = Input >- (comp2,"-r","-x") >|
            Input >- (comp1,"-r","-x") >|
            Input >- (comp3,"-r","-x") >|> (comp1,"-x","-x")

main :: IO ()
main = do putStrLn ascii_art
          let result = sequence1 #>> 's'
          putStrLn result
          writeFile "script.pl" result
          putStrLn "___________________THE_END____________________\n"

ascii_art = unlines [l1,l2,l3,l4,l5,l6]
l1 = " .----..                                .---.     "
l2 = " '---.  `------------------------------´  _  `.   "
l3 = "      \\          HASKELL/CROSS           / \\  : "
l4 = "      /   ----------------------------   \\_/  :  "
l5 = " .---´  ,'                            `.     .'   "
l6 = "  `----'                                `---'     "


